// This package kms implements a simplistic key managment system (kms) for
// Quantum Key Distribution Networks (QKDN) which is a simple emulated KMS. x
// It relies on the emulated quantum link out of the quantumlayer package

package kms

import (
	"crypto/rand"
	"encoding/binary"
	"io"
	"sync"
	"time"

	log "github.com/sirupsen/logrus"

	pbETSI "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsetsi"
	pbIC "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsintercom"
	"code.fbi.h-da.de/danet/proto-kms/quantumlayer"
	"github.com/google/uuid"
)

type Qkdnkms interface {
	//AddExternalNotifierGeneral(chan bool)   // used to indicate unspecific changes
	AddExternalNotifierQLE(chan uint32)     // used to indicate changes to specific Quantum Link Element (QLE)
	AddExternalNotifierKMSPeer(chan string) // used to indicate changes to specific KMSPeer
	AddQuantumElement() *QuantumElement
	GlobalKeyHandler(time.Duration) error
	AddPeer(kmsPeerSocket string, servingQLE *QuantumElement)
	RemovePeer(kmsPeerSocket string)
	FindPeerUuid(uuid.UUID) *kmsPeer
}

type qlElementLinkID int

// The general emulated KMS
type EKMS struct {
	kmsName                 string
	kmsUUID                 uuid.UUID
	qleMapMutex             sync.Mutex
	QuantumElements         map[uint32]*QuantumElement
	externalNotifierQLE     chan uint32
	kmsPeersMutex           sync.Mutex
	KmsPeers                map[string]*kmsPeer
	externalNotifierKMSPeer chan string
	pbETSI.UnimplementedKmsETSIServer
	pbIC.UnimplementedKmsTalkerServer
}

// Will keep information about the quantum elements that this EKMS is talking to
// This actually constitutes a quantum element with only a single link

/*
type QuantumElementInterface interface {
	GetQlID() qlElementId
}*/

type QuantumElement struct {
	QlID               uint32
	QuantumElementLink *quantumlayer.QuantumlayerEmuPRNG // contains information about the quantum links
	//key stores go here
	keyStoreLocal  *kmsKeyStore // the keys this local entity has produced and are ready to use
	keyStoreRemote *kmsKeyStore // the keys th remote entity (peer) has produced and are ready to use
}

func NewEKMS(kmsName string, kmsUUID uuid.UUID, logOutput io.Writer, logLevel log.Level, logInJson bool) (newEKMS *EKMS) {

	/*
	 * Setup logging
	 */
	//What level
	log.SetLevel(logLevel)
	// Where to send log out put
	log.SetOutput(logOutput)
	// and plain-text (standard) or json
	if !logInJson {
		log.SetFormatter(&log.TextFormatter{})
	} else {
		log.SetFormatter(&log.JSONFormatter{})
	}
	// print code function if level is set to Trace
	if logLevel == log.TraceLevel {
		log.SetReportCaller(true)
	} else {
		log.SetReportCaller(false)
	}

	return &EKMS{
		kmsName:                 kmsName,
		kmsUUID:                 kmsUUID,
		QuantumElements:         make(map[uint32]*QuantumElement),
		KmsPeers:                make(map[string]*kmsPeer),
		externalNotifierQLE:     nil, // just be surely set to nil!
		externalNotifierKMSPeer: nil, // just be surely set to nil!
	}
}

func (kms *EKMS) AddQuantumElement(kmsUDPAddrr string, logOutput io.Writer, logLevel log.Level, logInJson bool) *QuantumElement {

	//Get an emulated Quantumlayer
	ql := quantumlayer.NewQuantumlayerEmuPRNG(logOutput, logLevel, logInJson)
	ql.Configure(kmsUDPAddrr)
	ql.PowerOn()

	ksl := kmsKeyStore{
		keyStore: make(map[string]*kmsKSElement),
	}
	ksr := kmsKeyStore{
		keyStore: make(map[string]*kmsKSElement),
	}

	qle := QuantumElement{
		QuantumElementLink: ql,
		keyStoreLocal:      &ksl,
		keyStoreRemote:     &ksr,
	}

	// generate a ID for this quantum element that is unique locally
	var randError error
	qle.QlID, randError = kms.GenerateNewQleID()
	if randError != nil {
		log.Fatalf("%s: GenerateNewQleID: %s", kms.kmsName, randError)
		return nil
	}

	kms.QuantumElements[qle.QlID] = &qle

	return &qle
}

func (kms *EKMS) GlobalKeyHandler(waitTime time.Duration) error {

	// periodically walk through QuantumElements and retrieve their
	// - local key bulk buffer
	// - remote key bulk buffer
	// feed this into the corresponding key buffers of the kmss
	for {
		for currentQE := range kms.QuantumElements {
			log.Debugf("%s GlobalKeyHandler reading...\n", kms.kmsName)

			bulkKeysLocal, err := kms.QuantumElements[currentQE].QuantumElementLink.GetKeyBatchLocal()
			if err != nil {
				log.Errorf("%s failed to retrieve local bulkkeys with error %s", kms.kmsName, err)
			} else {
				// process bulkKeysLocal
				log.Debugf("%s produced %d bytes of key locally", kms.kmsName, bulkKeysLocal.BulkKeyLength)
				kms.QuantumElements[currentQE].keyStoreLocal.KeyChopper256Bit(&bulkKeysLocal)

			}

			bulkKeysRemote, err := kms.QuantumElements[currentQE].QuantumElementLink.GetKeyBatchPeer()
			if err != nil {
				log.Errorf("%s failed to retrieve remote bulkkeys with error %s", kms.kmsName, err)
			} else {
				// process bulkKeysRemote
				log.Debugf("%s received %d bytes of key from remote peer", kms.kmsName, bulkKeysRemote.BulkKeyLength)
				kms.QuantumElements[currentQE].keyStoreRemote.KeyChopper256Bit(&bulkKeysRemote)
			}
		}
		// TODO: Better approach required than a sleep timer!
		time.Sleep(waitTime)
	}
}

// This has a design flaw, as the generated ID is returned to the calling function and used there.
// However, when being used a potential other caller might received the same qlElementId
// TODO/XXX: This would be collision and must be eventually avoided
func (kms *EKMS) GenerateNewQleID() (uint32, error) {
	for { // this needs a condition to stop!
		// create buffer for uint32, so reserve 4 bytes
		buf := make([]byte, 4)

		// fill the buffer from rand
		_, err := rand.Read(buf)
		if err != nil {
			return 0, err
		}

		propopsedQlElementID := binary.BigEndian.Uint32(buf)

		// check if ID is already taken
		if kms.QuantumElements[propopsedQlElementID] == nil {
			return propopsedQlElementID, nil
		}
		//keep going....
	}
}

// TODO/XXX error handling
func (kms *EKMS) AddPeer(kmsPeerSocket string, servingQLE *QuantumElement) {
	//check if peer exists
	if _, there := kms.KmsPeers[kmsPeerSocket]; there {
		log.Errorf("Trying to add existing peer %s", kmsPeerSocket)
		return
	}
	peer := NewKmsPeer(servingQLE, kms.externalNotifierKMSPeer)
	peer.tcpSocketStr = kmsPeerSocket

	kms.kmsPeersMutex.Lock()
	kms.KmsPeers[kmsPeerSocket] = &peer
	kms.kmsPeersMutex.Unlock()

	go peer.PeerHandler(kms.kmsName)
}

// TODO/XXX error handling
func (kms *EKMS) RemovePeer(kmsPeerSocket string) {
	if _, there := kms.KmsPeers[kmsPeerSocket]; there {
		//peer.quit <- true
		delete(kms.KmsPeers, kmsPeerSocket)
		return
	}
	log.Errorf("%s: Can not find a peer with socket: %s", kms.kmsName, kmsPeerSocket)
	return
}

func (kms *EKMS) AddExternalNotifierQLE(in chan uint32) {
	kms.externalNotifierQLE = in
}

func (kms *EKMS) AddExternalNotifierKMSPeer(in chan string) {
	kms.externalNotifierKMSPeer = in
}

func (kms *EKMS) FindPeerUuid(lookup uuid.UUID) (peer *kmsPeer) {
	if kms.KmsPeers != nil {
		for _, peer = range kms.KmsPeers {
			if peer.id == lookup {
				return peer
			}
		}
	}

	return nil
}
