package kms

import (
	"context"
	"encoding/base64"
	"errors"
	"flag"
	"fmt"
	"net"
	"os/exec"

	log "github.com/sirupsen/logrus"

	pb "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsetsi"
	"google.golang.org/grpc"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"
)

var (
	etsiPort = flag.Int("port", 50900, "The server port")
)

type etsiServer struct {
	pb.UnimplementedKmsETSIServer
	handlingEkms *EKMS
	visitedKeys  map[*kmsKSElement]bool
}

func (es *etsiServer) ETSICapabilities(ctx context.Context, in *pb.ETSICapabilitiesRequest) (capReply *pb.ETSICapabilitiesReply, err error) {
	log.Debugf("Received: %v", in.GetMyKmsName())

	return &pb.ETSICapabilitiesReply{
		PeerKmsName: "whatever",
	}, nil
}

func (es *etsiServer) ETSIGetQuantumInterfaces(ctx context.Context, in *pb.ETSIKMSQuantumInterfaceListRequest) (qleReply *pb.ETSIKMSQuantumInterfaceListReply, err error) {
	qleList := make([]*pb.QuantumElementInfo, 1)

	// Walk through QuantumLayerInterfaces and return their information
	for _, qlWorks := range es.handlingEkms.QuantumElements {
		qleElement := pb.QuantumElementInfo{
			QleID:   uint64(qlWorks.QlID),
			UdpAddr: fmt.Sprintf("%s:%d", qlWorks.QuantumElementLink.GetLocalQLPort().IP.String(), qlWorks.QuantumElementLink.GetLocalQLPort().Port),
		}
		qleList = append(qleList, &qleElement)
	}
	return &pb.ETSIKMSQuantumInterfaceListReply{
		QlElementInfo: qleList,
	}, nil
}

func (es *etsiServer) ETSIAddKMSPeer(ctx context.Context, in *pb.ETSIKMSPeerRequest) (*pb.ETSIKMSPeerReply, error) {
	//determine the kms structure to call
	log.Debugf("AddKMSPeer called.")

	// Check first if KmsLocalQLEId is actually one of ours...
	qleID := in.KmsLocalQLEId
	servingQLE, _ := es.handlingEkms.QuantumElements[qleID]
	if servingQLE == nil {
		//no such element!
		err := errors.New(fmt.Sprintf("Unknown local quantum element with ID %d", qleID))

		return &pb.ETSIKMSPeerReply{}, err
	}

	es.handlingEkms.AddPeer(in.GetKmsPeerSocket(), servingQLE)

	return &pb.ETSIKMSPeerReply{
		KmsPeerName: es.handlingEkms.kmsName,
	}, nil
}

func (es *etsiServer) ETSIRemoveKMSPeer(ctx context.Context, in *pb.ETSIKMSPeerRequest) (*pb.ETSIKMSPeerReply, error) {
	//kms.RemovePeer(in.GetKmsPeerSocket())

	return &pb.ETSIKMSPeerReply{
		//KmsPeerName: kms.kmsName,
	}, nil
}

func (es *etsiServer) ETSIGetPeerList(ctx context.Context, in *pb.ETSIKMSPeerListRequest) (*pb.ETSIKMSPeerListReply, error) {

	ep := make([]*pb.ETSIKMSPeer, 2)

	r := new(pb.ETSIKMSPeerListReply)

	r.Peer = ep

	p := pb.ETSIKMSPeer{
		PeerName:   "Yo",
		PeerStatus: "bla",
	}

	r.Peer[0] = &p

	return r, nil
}

func (es *etsiServer) ETSIGetEncryptKeys256Bit(ctx context.Context, in *pb.ETSIGetEncryptKeys256BitRequest) (*pb.ETSIGetEncryptKeys256BitReply, error) {
	log.Printf("Received request for n=%d keys", in.GetAmount())

	var randomKey *kmsKSElement
	var err error
	// NOTE: change change change!
	for _, qe := range es.handlingEkms.QuantumElements {
		quantumElementLocalKeyStore := qe.keyStoreLocal
		randomKey, err = randomItemFromMap[string, *kmsKSElement](quantumElementLocalKeyStore.keyStore, es.visitedKeys)
		if err != nil {
			return nil, status.Errorf(codes.Internal, "%v", err)
		}
		es.visitedKeys[randomKey] = true
	}

	// TODO: Remove/move ssh-kind prototype code below
	keyAsString := base64.StdEncoding.EncodeToString(randomKey.key)
	// push the key to the encryptor via ssh
	ssh := "ssh"
	complexArg := fmt.Sprintf("(rc=$(sed \"12 c PresharedKey = %s\" /etc/wireguard/wg0.conf); echo \"$rc\" > /etc/wireguard/wg0.conf)", keyAsString)
	args := []string{"root@172.20.0.5", "-oStrictHostKeyChecking=no", complexArg}

	cmd := exec.Command(ssh, args...)
	err = cmd.Run()
	if err != nil {
		log.Errorf("could not execute ssh command with parameters")
		return nil, status.Errorf(codes.Internal, "%v", err)
	}

	// Construct any response
	return &pb.ETSIGetEncryptKeys256BitReply{
		KeyID: randomKey.keyID,
		Key:   randomKey.key,
	}, nil
}

func StartETSI(listenAddr string, callingKMS *EKMS) {
	flag.Parse()

	//lis, err := net.Listen("tcp", fmt.Sprintf(":%d", *etsiPort))
	lis, err := net.Listen("tcp", listenAddr)
	if err != nil {
		log.Fatalf("failed to listen: %v", err)
	}
	s := grpc.NewServer()
	pb.RegisterKmsETSIServer(s, &etsiServer{
		handlingEkms: callingKMS,
		visitedKeys:  make(map[*kmsKSElement]bool),
	})
	log.Infof("server listening at %v", lis.Addr())
	if err := s.Serve(lis); err != nil {
		log.Fatalf("failed to serve: %v", err)
	}
}
