package kms

import (
	"context"
	"encoding/base64"
	"fmt"
	"net"
	"os/exec"
	"time"

	log "github.com/sirupsen/logrus"

	pb "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsetsi"
	"github.com/google/uuid"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
)

type KmsPeerStatus int16

const (
	KmsPeerUp      KmsPeerStatus = iota // peer was seen up lately
	KmsPeerDown                         // peer not reachable anymore
	KmsPeerPending                      // peer was added and has to be succesfully contacted once
	KmsPeerUnknown                      // not known, not initialized
)

type kmsPeerInfo interface {
	GetKmsPeerStatus() KmsPeerStatus
	GetKmsPeerId() uuid.UUID
	GetKmsPeerQkdiId() uint32
}

type kmsPeer struct {
	externalNotifierKMSPeer chan string
	peerStatus              KmsPeerStatus
	servingQLE              *QuantumElement
	tcpSocket               net.TCPAddr // the IP address and TCP port (aka socket) of the kms peer
	tcpSocketStr            string      // string rep. of tcpSocket
	name                    string      // the name of the kms peer
	id                      uuid.UUID   // uuid of the peer
	quit                    chan bool   // cancel the peer goroutine
}

func NewKmsPeer(servQLE *QuantumElement, in chan string) (peer kmsPeer) {
	return kmsPeer{
		peerStatus:              KmsPeerUnknown,
		servingQLE:              servQLE,
		id:                      uuid.New(),
		externalNotifierKMSPeer: in,
		quit:                    make(chan bool),
	}
}

// Handles everything with respect to a specific KMS peer
func (ph *kmsPeer) PeerHandler(kmsName string) {
	log.Infof("%s started PeerHandler for %s:", kmsName, ph.tcpSocketStr)

	// contact peer
	newPeerConn, err := grpc.Dial(ph.tcpSocketStr, grpc.WithTransportCredentials(insecure.NewCredentials()))
	if err != nil {
		log.Errorf("%s: did not connect: %v", kmsName, err)
		ph.peerStatus = KmsPeerDown
		return
	}
	defer newPeerConn.Close()

	c := pb.NewKmsETSIClient(newPeerConn)

	// Contact the server and print out its response.
	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()
	r, err := c.ETSICapabilities(ctx, &pb.ETSICapabilitiesRequest{MyKmsName: kmsName})
	if err != nil {
		log.Errorf("%s: could not greet: %v", kmsName, err)
		ph.peerStatus = KmsPeerDown
		// Send notification about change
		if ph.externalNotifierKMSPeer != nil {
			ph.externalNotifierKMSPeer <- ph.tcpSocketStr
		}
	}

	// NOTE: If a session is created, we also request a encryption key for now
	//quantumElementRequest, err := c.ETSIGetQuantumInterfaces(ctx, &pb.ETSIKMSQuantumInterfaceListRequest{})
	//if err != nil {
	//	log.Println("Could not request quantum elements: ", err)
	//	ph.peerStatus = KmsPeerDown
	//	// Send notification about change
	//	if ph.externalNotifierKMSPeer != nil {
	//		ph.externalNotifierKMSPeer <- ph.tcpSocketStr
	//	}
	//	return
	//}

	//if quantumElementRequest.QlElementInfo != nil {
	//	peerQuantumElement := quantumElementRequest.QlElementInfo[0].QleID
	//} else {
	//	// return error?!
	//}

	// Works and peer moves to kmsPeerUp
	ph.peerStatus = KmsPeerUp

	encryptKeyRequest, err := c.ETSIGetEncryptKeys256Bit(ctx, &pb.ETSIGetEncryptKeys256BitRequest{Amount: 1})
	if err != nil {
		log.Errorf("%s: could not request a encryption key: %s", kmsName, err)
		ph.peerStatus = KmsPeerDown
		// Send notification about change
		if ph.externalNotifierKMSPeer != nil {
			ph.externalNotifierKMSPeer <- ph.tcpSocketStr
		}
		return
	}

	// check if key is in remoteKeyStore
	if key, ok := ph.servingQLE.keyStoreRemote.keyStore[encryptKeyRequest.KeyID]; ok {
		keyAsString := base64.StdEncoding.EncodeToString(key.key)
		log.Debugf("Agreed Key: %s", keyAsString)

		// TODO: Remove/move ssh-kind prototype code below
		// push the key to the encryptor via ssh
		ssh := "ssh"
		complexArg := fmt.Sprintf("(rc=$(sed \"12 c PresharedKey = %s\" /etc/wireguard/wg0.conf); echo \"$rc\" > /etc/wireguard/wg0.conf)", keyAsString)
		args := []string{"root@172.20.0.4", "-oStrictHostKeyChecking=no", complexArg}

		cmd := exec.Command(ssh, args...)
		err := cmd.Run()
		if err != nil {
			log.Errorf("%s: could not execute ssh command with parameters", kmsName)
			return
		}
	}

	// Send notification about change
	if ph.externalNotifierKMSPeer != nil {
		ph.externalNotifierKMSPeer <- ph.tcpSocketStr
	}

	log.Printf("Greeting: %s which is now in peerStatus %d", r.GetPeerKmsName(), ph.peerStatus)

	// NOTE: should be possible to cancel!
	// By now, do check only the liveliness of the peer, nothing else.
	//for {
	//	select {
	//	case <-ph.quit:
	//		return
	//	default:
	//		// Contact the server and print out its response.
	//		ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	//		defer cancel()
	//		_, err := c.ETSICapabilities(ctx, &pb.ETSICapabilitiesRequest{MyKmsName: kmsName})
	//		if err != nil {
	//			log.Printf("could not greet: %v", err)
	//			ph.peerStatus = KmsPeerDown
	//			// Send notification about change
	//			if ph.externalNotifierKMSPeer != nil {
	//				ph.externalNotifierKMSPeer <- ph.tcpSocketStr
	//			}
	//		}
	//		time.Sleep(30 * time.Second)
	//	}
	//}
}

func (ph *kmsPeer) GetKmsPeerStatus() KmsPeerStatus {
	return ph.peerStatus
}

func (ph *kmsPeer) GetKmsPeerId() uuid.UUID {
	return ph.id
}

func (ph *kmsPeer) GetKmsPeerQkdiId() uint32 {
	return uint32(ph.servingQLE.QlID)
}
